-------------------------------------------------------------------------------
--
-- Copyright (C) 2015
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, unb2b_board_lib, technology_lib, diag_lib, dp_lib, tech_jesd204b_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE unb2b_board_lib.unb2b_board_pkg.ALL;
USE unb2b_board_lib.unb2b_board_peripherals_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.lofar2_unb2b_adc_pkg.ALL;

ENTITY lofar2_unb2b_adc IS
  GENERIC (
    g_design_name      : STRING  := "lofar2_unb2b_adc";
    g_design_note      : STRING  := "UNUSED";
    g_technology       : NATURAL := c_tech_arria10_e1sg;
    g_buf_nof_data     : NATURAL := 8192;
    g_sim              : BOOLEAN := FALSE; --Overridden by TB
    g_sim_unb_nr       : NATURAL := 0;
    g_sim_node_nr      : NATURAL := 0;
    g_sim_model_ddr    : BOOLEAN := FALSE;
    g_stamp_date       : NATURAL := 0;  -- Date (YYYYMMDD) -- set by QSF
    g_stamp_time       : NATURAL := 0;  -- Time (HHMMSS)   -- set by QSF
    g_revision_id      : STRING  := "";  -- revision ID     -- set by QSF
    g_factory_image    : BOOLEAN := FALSE;
    g_protect_addr_range: BOOLEAN := FALSE
  );
  PORT (
    -- GENERAL
    CLK          : IN    STD_LOGIC; -- System Clock
    PPS          : IN    STD_LOGIC; -- System Sync
    WDI          : OUT   STD_LOGIC; -- Watchdog Clear
    INTA         : INOUT STD_LOGIC; -- FPGA interconnect line
    INTB         : INOUT STD_LOGIC; -- FPGA interconnect line

    -- Others
    VERSION      : IN    STD_LOGIC_VECTOR(c_unb2b_board_aux.version_w-1 DOWNTO 0);
    ID           : IN    STD_LOGIC_VECTOR(c_unb2b_board_aux.id_w-1 DOWNTO 0);
    TESTIO       : INOUT STD_LOGIC_VECTOR(c_unb2b_board_aux.testio_w-1 DOWNTO 0);
    
    -- I2C Interface to Sensors
    SENS_SC      : INOUT STD_LOGIC;
    SENS_SD      : INOUT STD_LOGIC;
  
    PMBUS_SC     : INOUT STD_LOGIC;
    PMBUS_SD     : INOUT STD_LOGIC;
    PMBUS_ALERT  : IN    STD_LOGIC := '0';

    -- 1GbE Control Interface
    ETH_CLK      : IN    STD_LOGIC;
    ETH_SGIN     : IN    STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 DOWNTO 0);
    ETH_SGOUT    : OUT   STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 DOWNTO 0);

    -- LEDs
    QSFP_LED     : OUT   STD_LOGIC_VECTOR(c_unb2b_board_tr_qsfp_nof_leds-1 DOWNTO 0);

     -- back transceivers (Note: numbered from 0)
    JESD204B_SERIAL_DATA       : IN    STD_LOGIC_VECTOR((c_unb2b_board_tr_jesd204b.bus_w*c_unb2b_board_tr_jesd204b.nof_bus)-1 downto 0); 
                                                  -- Connect to the BCK_RX pins in the top wrapper
    JESD204B_REFCLK            : IN    STD_LOGIC; -- Connect to BCK_REF_CLK pin in the top level wrapper
 
    -- jesd204b syncronization signals
    JESD204B_SYSREF            : IN    STD_LOGIC;
    JESD204B_SYNC              : OUT   STD_LOGIC_VECTOR((c_unb2b_board_tr_jesd204b.bus_w*c_unb2b_board_tr_jesd204b.nof_bus)-1 DOWNTO 0)
  );
END lofar2_unb2b_adc;


ARCHITECTURE str OF lofar2_unb2b_adc IS

  -- Revision parameters
  CONSTANT c_revision_select        : t_lofar2_unb2b_adc_config := func_sel_revision_rec(g_design_name);
  CONSTANT c_nof_streams_jesd204b   : NATURAL := c_revision_select.nof_streams_jesd204b; -- IP is set up for 12 streams
  CONSTANT c_nof_streams_db         : NATURAL := c_revision_select.nof_streams_db;       -- Streams of raw samples to record in db
  CONSTANT c_nof_streams_input      : NATURAL := c_revision_select.nof_streams_input;    -- Streams actually passed through for processing

  -- Firmware version x.y
  CONSTANT c_fw_version             : t_unb2b_board_fw_version := (1, 1);
  CONSTANT c_mm_clk_freq            : NATURAL := c_unb2b_board_mm_clk_freq_100M;
  CONSTANT c_lofar2_sample_clk_freq : NATURAL := 200 * 10**6;  -- alternate 160MHz. TODO: Use to check PPS

  -- Waveform Generator
  CONSTANT c_wg_buf_directory       : STRING := "data/";
  CONSTANT c_wg_buf_dat_w           : NATURAL := c_unb2b_board_peripherals_mm_reg_default.ram_diag_wg_dat_w;
  CONSTANT c_wg_buf_addr_w          : NATURAL := c_unb2b_board_peripherals_mm_reg_default.ram_diag_wg_adr_w;
  SIGNAL wg_out_ovr                 : STD_LOGIC_VECTOR(c_nof_streams_input-1 DOWNTO 0);    
  SIGNAL wg_out_val                 : STD_LOGIC_VECTOR(c_nof_streams_input-1 DOWNTO 0);    
  SIGNAL wg_out_data                : STD_LOGIC_VECTOR(c_nof_streams_input*c_wg_buf_dat_w-1 DOWNTO 0);    
  SIGNAL wg_out_sync                : STD_LOGIC_VECTOR(c_nof_streams_input-1 DOWNTO 0);    
  SIGNAL wg_sosi_arr                : t_dp_sosi_arr(c_nof_streams_input-1 DOWNTO 0);    
  SIGNAL mux_sosi_arr               : t_dp_sosi_arr(c_nof_streams_input-1 DOWNTO 0);         
  SIGNAL nxt_mux_sosi_arr           : t_dp_sosi_arr(c_nof_streams_input-1 DOWNTO 0);

  -- bsn monitor
  SIGNAL bsn_sosi_arr               : t_dp_sosi_arr(c_nof_streams_jesd204b-1 DOWNTO 0) := (others => c_dp_sosi_rst);

  -- System
  SIGNAL cs_sim                     : STD_LOGIC;
  SIGNAL xo_ethclk                  : STD_LOGIC;
  SIGNAL xo_rst                     : STD_LOGIC;
  SIGNAL xo_rst_n                   : STD_LOGIC;
  SIGNAL mm_clk                     : STD_LOGIC;
  SIGNAL mm_rst                     : STD_LOGIC := '0';
  
  SIGNAL st_rst                     : STD_LOGIC;
  SIGNAL st_clk                     : STD_LOGIC;
  SIGNAL st_pps                     : STD_LOGIC;

  SIGNAL dp_rst                     : STD_LOGIC;
  SIGNAL dp_clk                     : STD_LOGIC;

  -- PIOs
  SIGNAL pout_wdi                   : STD_LOGIC;

  -- WDI override
  SIGNAL reg_wdi_mosi               : t_mem_mosi;
  SIGNAL reg_wdi_miso               : t_mem_miso;

  -- PPSH
  SIGNAL reg_ppsh_mosi              : t_mem_mosi;
  SIGNAL reg_ppsh_miso              : t_mem_miso;
  
  -- UniBoard system info
  SIGNAL reg_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL reg_unb_system_info_miso   : t_mem_miso;
  SIGNAL rom_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL rom_unb_system_info_miso   : t_mem_miso;

  -- UniBoard I2C sens
  SIGNAL reg_unb_sens_mosi          : t_mem_mosi;
  SIGNAL reg_unb_sens_miso          : t_mem_miso;

  -- pm bus
  SIGNAL reg_unb_pmbus_mosi         : t_mem_mosi;
  SIGNAL reg_unb_pmbus_miso         : t_mem_miso;

  -- FPGA sensors
  SIGNAL reg_fpga_temp_sens_mosi     : t_mem_mosi;
  SIGNAL reg_fpga_temp_sens_miso     : t_mem_miso;
  SIGNAL reg_fpga_voltage_sens_mosi  : t_mem_mosi;
  SIGNAL reg_fpga_voltage_sens_miso  : t_mem_miso;

  -- eth1g
  SIGNAL eth1g_mm_rst               : STD_LOGIC;
  SIGNAL eth1g_tse_mosi             : t_mem_mosi;  -- ETH TSE MAC registers
  SIGNAL eth1g_tse_miso             : t_mem_miso;
  SIGNAL eth1g_reg_mosi             : t_mem_mosi;  -- ETH control and status registers
  SIGNAL eth1g_reg_miso             : t_mem_miso;
  SIGNAL eth1g_reg_interrupt        : STD_LOGIC;   -- Interrupt
  SIGNAL eth1g_ram_mosi             : t_mem_mosi;  -- ETH rx frame and tx frame memory
  SIGNAL eth1g_ram_miso             : t_mem_miso;

  -- EPCS read
  SIGNAL reg_dpmm_data_mosi         : t_mem_mosi;
  SIGNAL reg_dpmm_data_miso         : t_mem_miso;
  SIGNAL reg_dpmm_ctrl_mosi         : t_mem_mosi;
  SIGNAL reg_dpmm_ctrl_miso         : t_mem_miso;

  -- EPCS write
  SIGNAL reg_mmdp_data_mosi         : t_mem_mosi;
  SIGNAL reg_mmdp_data_miso         : t_mem_miso;
  SIGNAL reg_mmdp_ctrl_mosi         : t_mem_mosi;
  SIGNAL reg_mmdp_ctrl_miso         : t_mem_miso;

  -- EPCS status/control
  SIGNAL reg_epcs_mosi              : t_mem_mosi;
  SIGNAL reg_epcs_miso              : t_mem_miso;

  -- Remote Update
  SIGNAL reg_remu_mosi              : t_mem_mosi;
  SIGNAL reg_remu_miso              : t_mem_miso;

  -- JESD
  SIGNAL jesd204b_mosi              : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL jesd204b_miso              : t_mem_miso := c_mem_miso_rst;

  -- WG
  SIGNAL reg_wg_mosi_arr            : t_mem_mosi_arr(12-1 DOWNTO 0);
  SIGNAL reg_wg_miso_arr            : t_mem_miso_arr(12-1 DOWNTO 0);
  SIGNAL ram_wg_mosi_arr            : t_mem_mosi_arr(12-1 DOWNTO 0);
  SIGNAL ram_wg_miso_arr            : t_mem_miso_arr(12-1 DOWNTO 0);

  -- BSN MONITOR
  SIGNAL reg_bsn_monitor_input_mosi : t_mem_mosi;
  SIGNAL reg_bsn_monitor_input_miso : t_mem_miso;

  -- QSFP leds
  SIGNAL qsfp_green_led_arr         : STD_LOGIC_VECTOR(c_unb2b_board_tr_qsfp.nof_bus-1 DOWNTO 0);
  SIGNAL qsfp_red_led_arr           : STD_LOGIC_VECTOR(c_unb2b_board_tr_qsfp.nof_bus-1 DOWNTO 0);

  -- JESD signals
  SIGNAL ram_diag_data_buf_jesd_mosi    : t_mem_mosi;
  SIGNAL ram_diag_data_buf_jesd_miso    : t_mem_miso;
  SIGNAL reg_diag_data_buf_jesd_mosi    : t_mem_mosi;
  SIGNAL reg_diag_data_buf_jesd_miso    : t_mem_miso;
  SIGNAL diag_data_buf_snk_in_arr       : t_dp_sosi_arr(c_nof_streams_db-1 DOWNTO 0);
  SIGNAL jesd204b_rx_src_out_arr        : t_dp_sosi_arr(c_nof_streams_jesd204b-1 DOWNTO 0);
  SIGNAL jesd204b_frame_clk             : STD_LOGIC;

  -------------------------------------------------------------------------------
  -- DP sync checker / insert
  -------------------------------------------------------------------------------
  CONSTANT c_nof_clk_per_blk         : NATURAL := 1024;
  CONSTANT c_nof_blk_per_sync        : NATURAL := 800000;
  CONSTANT c_nof_clk_per_sync        : NATURAL := c_nof_blk_per_sync * 256;  -- = 800000 * 256
  CONSTANT c_bsn_sync_timeout        : NATURAL := (c_nof_clk_per_sync * 10)/8; -- *10/8 as margin



BEGIN

  -----------------------------------------------------------------------------
  -- General control function
  -----------------------------------------------------------------------------
  u_ctrl : ENTITY unb2b_board_lib.ctrl_unb2b_board
  GENERIC MAP (
    g_sim                => g_sim,
    g_technology         => g_technology,
    g_design_name        => g_design_name,
    g_design_note        => g_design_note,
    g_stamp_date         => g_stamp_date,
    g_stamp_time         => g_stamp_time, 
    g_revision_id        => g_revision_id, 
    g_fw_version         => c_fw_version,
    g_mm_clk_freq        => c_mm_clk_freq,
    g_eth_clk_freq       => c_unb2b_board_eth_clk_freq_125M,
    g_aux                => c_unb2b_board_aux,
    g_factory_image      => g_factory_image,
    g_protect_addr_range => g_protect_addr_range,
    g_dp_clk_use_pll     => FALSE
  )
  PORT MAP (
    -- Clock an reset signals
    cs_sim                   => cs_sim,
    xo_ethclk                => xo_ethclk,
    xo_rst                   => xo_rst,
    xo_rst_n                 => xo_rst_n,

    mm_clk                   => mm_clk,
    mm_rst                   => mm_rst,

    dp_rst                   => dp_rst,
    dp_clk                   => dp_clk,              -- Can be external 200MHz, or PLL generated
    dp_pps                   => st_pps,
    dp_rst_in                => st_rst,
    dp_clk_in                => jesd204b_frame_clk,
    
    -- Toggle WDI
    pout_wdi                 => pout_wdi,

    -- MM buses
    -- REMU
    reg_remu_mosi            => reg_remu_mosi,
    reg_remu_miso            => reg_remu_miso,

    -- EPCS read
    reg_dpmm_data_mosi       => reg_dpmm_data_mosi,
    reg_dpmm_data_miso       => reg_dpmm_data_miso,
    reg_dpmm_ctrl_mosi       => reg_dpmm_ctrl_mosi,
    reg_dpmm_ctrl_miso       => reg_dpmm_ctrl_miso,

    -- EPCS write
    reg_mmdp_data_mosi       => reg_mmdp_data_mosi,
    reg_mmdp_data_miso       => reg_mmdp_data_miso,
    reg_mmdp_ctrl_mosi       => reg_mmdp_ctrl_mosi,
    reg_mmdp_ctrl_miso       => reg_mmdp_ctrl_miso,

    -- EPCS status/control
    reg_epcs_mosi            => reg_epcs_mosi,
    reg_epcs_miso            => reg_epcs_miso,

    -- . Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,
    
    -- . System_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso, 
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 
    
    -- . UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso,    
    
    -- . FPGA sensors
    reg_fpga_temp_sens_mosi  => reg_fpga_temp_sens_mosi,
    reg_fpga_temp_sens_miso  => reg_fpga_temp_sens_miso,
    reg_fpga_voltage_sens_mosi  => reg_fpga_voltage_sens_mosi,
    reg_fpga_voltage_sens_miso  => reg_fpga_voltage_sens_miso,

    reg_unb_pmbus_mosi       => reg_unb_pmbus_mosi,
    reg_unb_pmbus_miso       => reg_unb_pmbus_miso,

    -- . PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso,
    
    -- eth1g
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,
        
    -- FPGA pins
    -- . General
    CLK                      => jesd204b_frame_clk,
    PPS                      => PPS,
    WDI                      => WDI,
    INTA                     => INTA,
    INTB                     => INTB,
    -- . Others
    VERSION                  => VERSION,
    ID                       => ID,
    TESTIO                   => TESTIO,
    -- . I2C Interface to Sensors
    SENS_SC                  => SENS_SC,
    SENS_SD                  => SENS_SD,
    -- PM bus
    PMBUS_SC                 => PMBUS_SC,
    PMBUS_SD                 => PMBUS_SD,
    PMBUS_ALERT              => PMBUS_ALERT,

    -- . 1GbE Control Interface
    ETH_clk                  => ETH_CLK,
    ETH_SGIN                 => ETH_SGIN,
    ETH_SGOUT                => ETH_SGOUT
  );

  -----------------------------------------------------------------------------
  -- MM master
  -----------------------------------------------------------------------------
  u_mmm : ENTITY work.mmm_lofar2_unb2b_adc
  GENERIC MAP (
    g_sim         => g_sim,
    g_sim_unb_nr  => g_sim_unb_nr,
    g_sim_node_nr => g_sim_node_nr
   )
  PORT MAP(  
    mm_rst                   => mm_rst,
    mm_clk                   => mm_clk,       

    -- PIOs
    pout_wdi                 => pout_wdi,

    -- Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,

    -- system_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso,
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 

    -- UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso, 

    reg_unb_pmbus_mosi       => reg_unb_pmbus_mosi,
    reg_unb_pmbus_miso       => reg_unb_pmbus_miso,
 
    -- FPGA sensors
    reg_fpga_temp_sens_mosi  => reg_fpga_temp_sens_mosi,
    reg_fpga_temp_sens_miso  => reg_fpga_temp_sens_miso,
    reg_fpga_voltage_sens_mosi  => reg_fpga_voltage_sens_mosi,
    reg_fpga_voltage_sens_miso  => reg_fpga_voltage_sens_miso,

    -- PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso, 
  
    -- eth1g
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,

    -- EPCS read
    reg_dpmm_data_mosi       => reg_dpmm_data_mosi,
    reg_dpmm_data_miso       => reg_dpmm_data_miso,
    reg_dpmm_ctrl_mosi       => reg_dpmm_ctrl_mosi,
    reg_dpmm_ctrl_miso       => reg_dpmm_ctrl_miso,

    -- EPCS write
    reg_mmdp_data_mosi       => reg_mmdp_data_mosi,
    reg_mmdp_data_miso       => reg_mmdp_data_miso,
    reg_mmdp_ctrl_mosi       => reg_mmdp_ctrl_mosi,
    reg_mmdp_ctrl_miso       => reg_mmdp_ctrl_miso,

    -- EPCS status/control
    reg_epcs_mosi            => reg_epcs_mosi,
    reg_epcs_miso            => reg_epcs_miso,

    -- Remote Update
    reg_remu_mosi            => reg_remu_mosi,
    reg_remu_miso            => reg_remu_miso,

    --
    ram_diag_data_buf_jesd_mosi => ram_diag_data_buf_jesd_mosi,
    ram_diag_data_buf_jesd_miso => ram_diag_data_buf_jesd_miso,
    reg_diag_data_buf_jesd_mosi => reg_diag_data_buf_jesd_mosi,
    reg_diag_data_buf_jesd_miso => reg_diag_data_buf_jesd_miso,

    -- BSN Monitor
    reg_bsn_monitor_input_mosi => reg_bsn_monitor_input_mosi,
    reg_bsn_monitor_input_miso => reg_bsn_monitor_input_miso,

    -- WGs 
    reg_wg_mosi_arr          => reg_wg_mosi_arr,
    reg_wg_miso_arr          => reg_wg_miso_arr,
    ram_wg_mosi_arr          => ram_wg_mosi_arr,
    ram_wg_miso_arr          => ram_wg_miso_arr,
 
    -- Jesd ip status/control
    jesd204b_mosi            => jesd204b_mosi,
    jesd204b_miso            => jesd204b_miso
  );

  
  -----------------------------------------------------------------------------
  -- JESD204B IP (ADC Handler)
  -----------------------------------------------------------------------------
  
  u_jesd204b: ENTITY tech_jesd204b_lib.tech_jesd204b 
  GENERIC MAP(
    g_sim                => g_sim,                
    g_nof_channels       => c_nof_streams_jesd204b,
    g_nof_syncs          => c_nof_streams_jesd204b/3        -- Three ADCs per RCU share a sync
  )
  PORT MAP(
    jesd204b_refclk      => JESD204B_REFCLK,   
    jesd204b_sysref      => JESD204B_SYSREF,   
    jesd204b_sync_n_arr  => JESD204B_SYNC,   
  
    rx_src_out_arr       => jesd204b_rx_src_out_arr,          
    jesd204b_frame_clk   => jesd204b_frame_clk,          
  
    -- MM
    mm_clk               => mm_clk,           
    mm_rst               => mm_rst,           
  
    jesd204b_mosi        => jesd204b_mosi,         
    jesd204b_miso        => jesd204b_miso,         
  
     -- Serial
    serial_tx_arr        => open,
    serial_rx_arr        => JESD204B_SERIAL_DATA(c_nof_streams_jesd204b-1 downto 0)
  );


  gen_jesd_mon_in : FOR i IN 0 TO c_nof_streams_db-1 GENERATE
    diag_data_buf_snk_in_arr(i).data(15 downto 0) <= jesd204b_rx_src_out_arr(i).data(15 downto 0);
    diag_data_buf_snk_in_arr(i).valid <= jesd204b_rx_src_out_arr(i).valid;
    diag_data_buf_snk_in_arr(i).sop   <= '0';
    diag_data_buf_snk_in_arr(i).eop   <= '0';
    diag_data_buf_snk_in_arr(i).err   <= (OTHERS=>'0');
  END GENERATE;


  -----------------------------------------------------------------------------
  -- Diagnostic Data Buffer (Records 8192 raw ADC samples after the PPS)
  -----------------------------------------------------------------------------

  u_diag_data_buffer : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (
    g_technology   => g_technology,
    g_nof_streams  => c_nof_streams_db,
    g_data_w       => 16,
    g_buf_nof_data => g_buf_nof_data,
    g_buf_use_sync => TRUE, -- when TRUE start filling the buffer at the in_sync, else after the last word was read
    g_use_rx_seq   => FALSE
  )
  PORT MAP (
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => st_rst,
    dp_clk            => jesd204b_frame_clk,

    ram_data_buf_mosi => ram_diag_data_buf_jesd_mosi,
    ram_data_buf_miso => ram_diag_data_buf_jesd_miso,
    reg_data_buf_mosi => reg_diag_data_buf_jesd_mosi,
    reg_data_buf_miso => reg_diag_data_buf_jesd_miso,

    in_sosi_arr       => diag_data_buf_snk_in_arr,
    in_sync           => st_pps
  );

  -----------------------------------------------------------------------------
  -- WG (Test Signal Generator)
  -----------------------------------------------------------------------------

  gen_wg : FOR I IN 0 TO c_nof_streams_input-1 GENERATE
    u_sp : ENTITY diag_lib.mms_diag_wg_wideband
    GENERIC MAP (
      g_cross_clock_domain => TRUE,
      g_buf_dir            => c_wg_buf_directory,

      -- Wideband parameters
      g_wideband_factor    => 1, 

      -- Basic WG parameters, see diag_wg.vhd for their meaning
      g_buf_dat_w          => c_wg_buf_dat_w,
      g_buf_addr_w         => c_wg_buf_addr_w,
      g_calc_support       => TRUE,
      g_calc_gain_w        => 1,
      g_calc_dat_w         => c_wg_buf_dat_w
    )
    PORT MAP (
      -- Memory-mapped clock domain
      mm_rst              => mm_rst,
      mm_clk              => mm_clk,

      reg_mosi            => reg_wg_mosi_arr(I),
      reg_miso            => reg_wg_miso_arr(I),

      buf_mosi            => ram_wg_mosi_arr(I),
      buf_miso            => ram_wg_miso_arr(I),

      -- Streaming clock domain
      st_rst              => dp_rst,
      st_clk              => jesd204b_frame_clk,
      st_restart          => st_pps,

      out_ovr             => wg_out_ovr(I downto I),
      out_val             => wg_out_val(I downto I),
      out_dat             => wg_out_data((I+1)*c_wg_buf_dat_w-1 downto I*c_wg_buf_dat_w),
      out_sync            => wg_out_sync(I downto I)
    );

    wg_sosi_arr(I).err(0) <= wg_out_ovr(I);
    wg_sosi_arr(I).valid <= wg_out_val(I);
    wg_sosi_arr(I).data(c_wg_buf_dat_w-1 downto 0) <= wg_out_data((I+1)*c_wg_buf_dat_w-1 downto I*c_wg_buf_dat_w);
    wg_sosi_arr(I).sync <= wg_out_sync(I);

  END GENERATE;



  -----------------------------------------------------------------------------
  -- ADC/WG Mux (Input Select)
  -----------------------------------------------------------------------------
  
  gen_mux : FOR I IN 0 TO c_nof_streams_input-1 GENERATE
    p_sosi : PROCESS(jesd204b_rx_src_out_arr, wg_sosi_arr)
    BEGIN
      -- Valid is forced to '1' here for dp_shiftram.
      nxt_mux_sosi_arr(I).valid <= '1';

      -- Default use the ADUH data
      nxt_mux_sosi_arr(I).data  <= jesd204b_rx_src_out_arr(I).data;
      IF wg_sosi_arr(I).valid='1' THEN
        -- Valid WG data overrules ADUH data
        nxt_mux_sosi_arr(I).data <= wg_sosi_arr(I).data;
      END IF;
    END PROCESS;
  END GENERATE;

  p_reg_mux : PROCESS(st_rst, jesd204b_frame_clk)
  BEGIN
    IF st_rst='1' THEN
      mux_sosi_arr  <= (OTHERS=>c_dp_sosi_rst);
    ELSIF rising_edge(jesd204b_frame_clk) THEN
      mux_sosi_arr  <= nxt_mux_sosi_arr;
    END IF;
  END PROCESS;


  ---------------------------------------------------------------------------------------
  -- BSN monitor (Block Checker)
  ---------------------------------------------------------------------------------------
  u_bsn_monitor : ENTITY dp_lib.mms_dp_bsn_monitor
  GENERIC MAP (
    g_nof_streams        => c_nof_streams_jesd204b,
    g_sync_timeout       => c_bsn_sync_timeout,
    g_bsn_w              => 51, --c_apertif_bsn_w,
    g_log_first_bsn      => FALSE
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,
    reg_mosi    => reg_bsn_monitor_input_mosi,
    reg_miso    => reg_bsn_monitor_input_miso,

    -- Streaming clock domain
    dp_rst      => dp_rst,
    dp_clk      => dp_clk,
    in_siso_arr => (OTHERS=>c_dp_siso_rdy),
    in_sosi_arr => bsn_sosi_arr
  );

  -- only connect the channels actually used

  gen_bsn_monitor_inputs : FOR I IN 0 TO c_nof_streams_input-1 GENERATE
    bsn_sosi_arr(I) <= mux_sosi_arr(I);
  END GENERATE;

END str;
